import argparse
import subprocess
import shutil
import sys
import glob
import os
import time
import json
import threading
from functools import wraps
from datetime import datetime

third_party_module_check_script = '/usr/libexec/upgrade-oc/check_user_installed_modules.sh'
system_service_check_script = '/usr/libexec/upgrade-oc/check_system_service.sh'
backup_script = '/usr/libexec/upgrade-oc/backup_restore.sh'

log_agent = '/usr/libexec/upgrade-oc/migration-log'

tmp_log_dir = '/var/log/upgrade-oc/'
leapp_log_dir = '/var/log/leapp/'

single_task_flag = os.path.join(tmp_log_dir, 'upgrade-oc-single-task.flag')
single_task_post_flag = os.path.join(tmp_log_dir, 'upgrade-oc-single-task-post.flag')

# Upgrade-oc log files
module_check_log = os.path.join(tmp_log_dir, 'module_check.log')
service_status_log = os.path.join(tmp_log_dir, 'service_status.txt')
summary_log = os.path.join(tmp_log_dir, 'summary.json')
backup_log = os.path.join(tmp_log_dir, 'backup.log')
postcheck_log = os.path.join(tmp_log_dir, 'postcheck.log')
user_app_check_log = os.path.join(tmp_log_dir, 'user_app_check.log')
upgrade_oc_log = os.path.join(tmp_log_dir, 'upgrade-oc.log')
leapp_upgrade_brief_log = os.path.join(tmp_log_dir, 'leapp-upgrade-brief.log')
leapp_preupgrade_brief_log = os.path.join(tmp_log_dir, 'leapp-preupgrade-brief.log')

# leapp logs
leapp_preupgrade_log = os.path.join(leapp_log_dir, 'leapp-preupgrade.log')
leapp_upgrade_log = os.path.join(leapp_log_dir, 'leapp-upgrade.log')
leapp_report_json = os.path.join(leapp_log_dir, 'leapp-report.json')

supported_versions = {
    'Tencent tlinux 2.2 (Final)': '2.2',
    'TencentOS Server 2.4': '2.4',
    'Tencent tlinux 2.6': '2.6',
    'TencentOS Server 4.2': '4.2',
    'TencentOS Server 4.0': '4.0',
    'TencentOS Server 3.2 (Final)': '3.2',
    'TencentOS Server 3.1 (Final)': '3.1',
}

supported_path = {
    "2.2": ["4.2","3.3","4.0","3.2"],
    "2.4": ["4.2","3.3","4.0","3.1"],
    "2.6": ["4.2","4.0","3.2"]
}

def log(msg):
    print(msg)

    if not os.path.exists(tmp_log_dir):
        os.makedirs(tmp_log_dir)

    timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    with open(upgrade_oc_log, 'a') as f:
        f.write("[{}] {}\n".format(timestamp, msg))

def single_task_decorator(flag_file=single_task_flag):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # 检查标志文件是否存在
            if os.path.exists(flag_file):
                raise RuntimeError("Another task is already running.")
            
            # 创建标志文件
            with open(flag_file, 'w') as f:
                f.write('Task is running.')
            
            try:
                # 执行被装饰的函数
                result = func(*args, **kwargs)
            finally:
                # 删除标志文件
                if os.path.exists(flag_file):
                    os.remove(flag_file)

            return result
        return wrapper
    return decorator


def check_upgrade_path(version):
    current_version = get_pretty_name()

    if current_version not in supported_versions:
        log("Current version {} is not supported.".format(current_version))
        return False

    if version not in supported_path.get(supported_versions[current_version], []):
        log("Current version {} is not supported for upgrade to {}.".format(current_version, version))
        return False

    return True


def prepare_env(version):
    log("Executing prepare_env...")
    success = True

    if not check_upgrade_path(supported_versions[version]):
        return False

    try:
        download_leapp(supported_versions[version])
        download_rpm_check()

        if not os.path.exists(tmp_log_dir):
            os.makedirs(tmp_log_dir)

    except Exception as e:
        log("Failed to prepare environment: %s" % str(e))
        success = False
    
    log("Prepare_env completed successfully.")
    return success

def download_leapp(version):
    if version == "4.0" or version == "4.2":
        subprocess.call(["yum", "install", "-y", "leapp-upgrade-ts2tots4"])
    elif version == "3.1" or version == "3.2":
        subprocess.call(["yum", "install", "-y", "leapp-upgrade-ts2to3"])
    else:
        log("Unknown version: %s" % version)
        sys.exit(1)

def download_rpm_check():
    subprocess.call(["yum", "install", "-y", "rpm-check"])

def module_check(force, uuid):
    log("Executing module_check...")

    success = True

    try:
        # 执行 check_user_installed_modules.sh 脚本并将输出重定向到 module_check.log
        with open(module_check_log, 'w') as module_log:
            result = subprocess.call(third_party_module_check_script, stdout=module_log, stderr=subprocess.STDOUT)
            if result != 0:
                log("Module check failed.")
                if not force:
                    success = False
            else:
                log("Module check completed successfully.")
    except Exception as e:
        log("An error occurred while executing module_check: %s" % str(e))
        if not force:
            success = False
    finally:
        # 检查 module_check.log 文件是否存在
        if os.path.exists(module_check_log):
            # 上传 module_check.log 文件
            upload_log(module_check_log, uuid, "precheck")
        
        return success

def system_service_check(force, uuid):
    log("Executing system_service_check...")

    success = True

    try:
        # 执行 check_system_service.sh record 命令并将输出重定向到 service_status.txt
        result = subprocess.call([system_service_check_script, 'record', service_status_log], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        if result != 0:
            log("Service status check failed.")
            if not force:
                success = False
        else:
            log("Service status check completed successfully.")
    except Exception as e:
        log("An error occurred while executing service status check: %s" % str(e))
        if not force:
            success = False
    finally:
        # 检查 service_status.txt 文件是否存在
        if os.path.exists(service_status_log):
            # 上传 service_status.txt 文件
            upload_log(service_status_log, uuid, "precheck")
        return success

def get_pretty_name():
    file_path='/etc/os-release'

    try:
        with open(file_path, 'r') as file:
            for line in file:
                if line.startswith('PRETTY_NAME='):
                    # 去掉前缀 'PRETTY_NAME=' 并去掉引号
                    pretty_name = line[len('PRETTY_NAME='):].strip().strip('"')
                    return pretty_name
    except FileNotFoundError:
        log("File {} not found.".format(file_path))
    except Exception as e:
        log("An error occurred: {}".format(str(e)))

    return None

def rpm_check(userapp, serverip, target, force, uuid):
    log("Executing rpm_check...")

    success = True

    try:
        # 执行 rpm-check 命令
        if userapp and serverip:
            # 4.2 4.0, we all consider them as 4, 2.2 2.6 as 2
            current_ver = supported_versions[get_pretty_name()].split('.')[0]
            target_version = supported_versions[target].split('.')[0]
            serverip = serverip + '/ts' + current_ver + 'tots' + target_version

            rpm_check_cmd = ['rpm-check', '--checkabi', '--check-local', userapp, '--search-remote', serverip, '--with-symbol']
            result = subprocess.call(rpm_check_cmd)
            if result != 0:
                log("RPM check failed.")
                if not force:
                    success = False
            else:
                log("RPM check completed successfully.")
        else:
            log("Skipping RPM check due to missing userapp or serverip.")
    except Exception as e:
        log("An error occurred while executing rpm-check: %s" % str(e))
        if not force:
            success = False
    finally:
        # 检查 summary.log 文件是否存在
        if os.path.exists("./summary.json"):
            # 上传 summary.log 文件
            shutil.move("./summary.json", summary_log)
            upload_log(summary_log, uuid, "precheck")
        return success

def leapp_check(backup_dir, exclude_dirs, force, uuid):
    log("Executing leapp_check...")

    success = True

    try:
        # 执行 leapp --preupgrade 命令
        command = ['leapp', 'preupgrade', '--tencentos']
        if uuid:
            # 构建 migration-log 命令
            log_command = [log_agent, 'log-file', '--filepath={}'.format(leapp_preupgrade_brief_log), '--uuid={}'.format(uuid), '--status=begin', '--stage=precheck']

            result = subprocess.call(" ".join(command) + " | tee {} | ".format(leapp_preupgrade_brief_log) + " ".join(log_command), shell=True)
        else:
            with open(leapp_preupgrade_brief_log, 'w') as preupgrade_log:
                result = subprocess.call(command, stdout=preupgrade_log, stderr=preupgrade_log)

        if result != 0:
            log("Leapp preupgrade failed.")
            success = False
        else:
            log("Leapp preupgrade completed successfully.")

        ret = backup("check", backup_dir, exclude_dirs, uuid)
        if not force and not ret:
            success = False
    except Exception as e:
        log("An error occurred while executing leapp-check: %s" % str(e))
        if not force:
            success = False
    finally:
        # 通知服务器滚动日志结束
        log_agent_mark(uuid, "precheck", "finish")

        # 检查 leapp-preupgrade-brief.log 文件是否存在
        #if os.path.exists(leapp_preupgrade_log):
        #    # 上传 leapp-preupgrade-brief.log 文件
        #    upload_log(leapp_preupgrade_brief_log, uuid, "precheck")
        
        if os.path.exists(leapp_report_json):
            upload_log(leapp_report_json, uuid, "precheck")

        return success

@single_task_decorator()
def precheck(backup_dir, exclude_dirs, userapp=None, serverip=None, target=None, force=False, uuid=None):
    log("Executing precheck...")

    any_failure = False

    try:
        if not module_check(force, uuid):
            any_failure = True

        if not system_service_check(force, uuid):
            any_failure = True

        if userapp and serverip and not rpm_check(userapp, serverip, target, force, uuid):
            any_failure = True

        if not leapp_check(backup_dir, exclude_dirs, force, uuid):
            any_failure = True

    except Exception as e:
        log("An unexpected error occurred during precheck: {}".format(e))
        any_failure = True

    finally:
        if any_failure:
            log("Precheck completed with failures.")
            return False

        log("Precheck completed successfully.")
        return True

def backup(action, backup_dir, exclude_dirs, uuid):
    log("Executing backup...")

    success = True

    if not backup_dir or not exclude_dirs:
        log("Backup directory or exclude directories not specified, skipping backup.")
        return success

    try:
        # 构建 backup_restore.sh 命令
        exclude_dirs_str = ','.join(exclude_dirs)
        backup_cmd = [
            backup_script,
            '-a', action,
            '-d', backup_dir,
            '-e', exclude_dirs_str,
        ]

        # 执行 backup_restore.sh 命令
        with open(backup_log, 'a') as backuplog:
            ret = subprocess.call(backup_cmd, stdout=backuplog, stderr=backuplog)
            if ret != 0:
                log("Backup failed with return code {}".format(ret))
                success = False
            else:
                log("Backup completed successfully.")

    except subprocess.CalledProcessError as e:
        log("An error occurred during backup: {}".format(e))
        success = False
    finally:
        # 检查 backup.log是否存在
        if action == "backup" and os.path.exists(backup_log):
            # not upload here, because backup.log may be larger than tms channel.
            pass
            #upload_log(backup_log, uuid, "upgrade")

        return success


@single_task_decorator()
def upgrade(version, uuid):
    log("Executing upgrade...")

    success = True

    try:
        # 构建 leapp upgrade 命令
        command = ['leapp', 'upgrade', '--tencentos']
        if version and version in supported_versions.keys():
            command.extend(['--target', supported_versions[version]])

        ret = 0

        # 执行 leapp upgrade 命令
        if uuid:
            # 构建 migration-log 命令
            log_command = [log_agent, 'log-file', '--filepath={}'.format(leapp_upgrade_brief_log), '--uuid={}'.format(uuid), '--status=begin', '--stage=upgrade']

            ret = subprocess.call(" ".join(command) + " | tee {} | ".format(leapp_upgrade_brief_log) + " ".join(log_command), shell=True)
        else:
            with open(leapp_upgrade_brief_log, 'w') as upgrade_brief_log:
                ret = subprocess.call(command, stdout=upgrade_brief_log, stderr=upgrade_brief_log)

        if ret != 0:
            log("Upgrade failed.")
            success = False
        else:
            log("Upgrade completed successfully.")
    except subprocess.CalledProcessError as e:
        log("An error occurred during upgrade: {}".format(e))
        success = False
    finally:
        ## finish log report
        log_agent_mark(uuid, "upgrade", "finish")

        if os.path.exists(leapp_report_json):
            upload_log(leapp_report_json, uuid, "upgrade")

        # 如果成功，重启进入新系统
        if success:
            subprocess.call("reboot")

        return success

@single_task_decorator(single_task_post_flag)
def postcheck(commands=None, uuid=None):
    log("Executing postcheck...")

    try:
        # 执行用户提供的命令
        if commands:
            for command in commands.split(','):
                log("Running command: {}".format(command))
                with open(user_app_check_log, 'a') as user_log:
                    result = subprocess.call(command, stdout=user_log, stderr=user_log, shell=True)
                if result != 0:
                    log("Command '{}' failed with return code {}.".format(command, result))
                    return False

        # 执行 check_system_service.sh compare 命令
        log("Running system service comparison check...")
        with open(postcheck_log, 'w') as service_log:
            subprocess.call([system_service_check_script, 'compare', service_status_log], stdout=service_log, stderr=service_log)
        log("System service comparison check completed successfully.")
    except subprocess.CalledProcessError as e:
        log("An error occurred during system service comparison check: {}".format(e))
        return False
    finally:
        if os.path.exists(user_app_check_log):
            upload_log(user_app_check_log, uuid, "postcheck")
        
        if os.path.exists(postcheck_log):
            upload_log(postcheck_log, uuid, "postcheck")

    log("Postcheck completed successfully.")
    return True

@single_task_decorator(single_task_post_flag)
def clean_logs():
    if os.path.exists(leapp_log_dir):
        shutil.rmtree(leapp_log_dir)
        log("Directory {} has been deleted.".format(leapp_log_dir))
    else:
        log("Directory {} does not exist.".format(leapp_log_dir))

    if os.path.exists(tmp_log_dir):
        shutil.rmtree(tmp_log_dir)
        print("Directory {} has been deleted.".format(tmp_log_dir))
    else:
        print("Directory {} does not exist.".format(tmp_log_dir))

@single_task_decorator(single_task_post_flag)
def clean_rpms():
    log("Executing clean...")

    # 卸载软件包
    packages = [
        'rpm-check',
        'rpm-check-tools',
        'leapp',
        'leapp-deps',
        'leapp-repository-deps',
    ]

    for package in packages:
        try:
            log("Uninstalling package: {}".format(package))
            subprocess.call(['yum', 'remove', '-y', package])
            log("Package {} uninstalled successfully.".format(package))
        except subprocess.CalledProcessError as e:
            log("An error occurred while uninstalling package {}: {}".format(package, e))
            return False

    print("Clean completed successfully.")
    return True


@single_task_decorator(single_task_post_flag)
def clean_backup(backup_dir):
    # 删除备份
    if os.path.exists(backup_dir):
        try:
            shutil.rmtree(backup_dir)
            log("Deleted backup directory: {}".format(backup_dir))
        except Exception as e:
            log("An error occurred while deleting backup directory {}: {}".format(backup_dir, e))
            
            return False
    
    log("Backup deleted successfully.")
    return True


def link_service_to_default_target_wants(service_file_path):
    """
    将指定的 service 文件链接到 default.target.wants 目录下。
    如果目录不存在，则创建该目录。

    :param service_file_path: 要链接的 service 文件的路径
    """
    # 定义 default.target.wants 目录的路径
    default_target_wants_dir = '/etc/systemd/system/default.target.wants'

    # 检查 default.target.wants 目录是否存在，如果不存在则创建
    if not os.path.exists(default_target_wants_dir):
        os.makedirs(default_target_wants_dir)
        log("Created directory: {}".format(default_target_wants_dir))

    # 获取 service 文件的文件名
    service_file_name = os.path.basename(service_file_path)

    # 定义目标路径
    target_path = os.path.join(default_target_wants_dir, service_file_name)

    # 检查是否已经存在同名的符号链接或文件
    if os.path.exists(target_path):
        if os.path.islink(target_path):
            os.remove(target_path)
        else:
            raise FileExistsError("File '{}' already exists and is not a symlink.".format(target_path))

    # 创建符号链接
    os.symlink(service_file_path, target_path)
    log("Linked '{}' to '{}'".format(service_file_path, target_path))


def create_systemd_service(userscript, backup_dir, uuid, target):

    service_name = "upgrade_oc_postcheck"

    if supported_versions[target].startswith("4."):
        version_dir = "ts2to4"
    else:
        version_dir = "ts3to3"

    leapp_bundle_dir = "/usr/share/leapp-repository/repositories/system_upgrade/{}/files/bundled-rpms/".format(version_dir)
    upgrade_oc_tool_dir = "/usr/libexec/upgrade-oc/"

    if not copy_new_version_to_leapp_dir(upgrade_oc_tool_dir, leapp_bundle_dir, "upgrade-oc*.rpm"):
        log("Failed to copy new version of upgrade-oc to leapp directory.")
        return False

    # 定义服务单元文件内容
    service_content = """
[Unit]
Description=Custom Upgrade Service
After=network.target leapp_resume.service
DefaultDependencies=no

[Service]
Type=simple
ExecStart=upgrade-oc --postcheck --userscript '{}' --backup-dir {} --uuid {}
Restart=on-failure
""".format(userscript, backup_dir, uuid)

    # 服务单元文件路径
    service_file_path = "/etc/systemd/system/{}.service".format(service_name)

    try:
        # 将服务单元文件内容写入文件
        with open(service_file_path, 'w') as service_file:
            service_file.write(service_content)

        # 启用服务，使其开机自启动
        link_service_to_default_target_wants(service_file_path)

        # 重新加载 systemd 配置
        subprocess.call(['systemctl', 'daemon-reload'])
    except Exception as e:
        log("Failed to create service {}: {}".format(service_name, str(e)))
        return False
    finally:
        log("Service {} created and enabled successfully.".format(service_name))
        return True

def copy_new_version_to_leapp_dir(source_dir, destination_dir, pattern):
    # 确保目标目录存在
    if not os.path.exists(destination_dir):
        os.makedirs(destination_dir)

    # 查找源目录中的 RPM 文件
    rpm_files = glob.glob(os.path.join(source_dir, pattern))

    if not rpm_files:
        log("No RPM files found in the source directory.")
        return False

    # 复制每个 RPM 文件到目标目录
    for rpm_file in rpm_files:
        shutil.copy(rpm_file, destination_dir)
        log("Copied {} to {}".format(rpm_file, destination_dir))
    
    return True

def is_service_active(service_name):
    try:
        result = subprocess.Popen(['systemctl', 'is-active', service_name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = result.communicate()
        return stdout.strip() == 'active'
    except Exception as e:
        log("Error checking service {}: {}".format(service_name, e))
        return False

def is_process_running(process_name):
    try:
        result = subprocess.Popen(['pgrep', '-f', process_name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = result.communicate()
        return result.returncode == 0
    except Exception as e:
        log("Error checking process {}: {}".format(process_name, e))
        return False

def wait_for_tms_tagent(max_fail=12):
    fail_time = 0
    while fail_time < max_fail:
        if (is_process_running('tmanager-service') and
            is_process_running('tdsp-taskchannel') and
            is_process_running('tagentV1.0')):
            log("All conditions met. Continuing...")
            # wait 2s for authorization
            time.sleep(2)
            return True
        else:
            log("Conditions not met. Waiting...")
            time.sleep(5)  # Wait for 5 seconds before checking again
            fail_time += 1

    log("Timeout reached. Conditions not met.")
    return False

def upload_logs_dir(uuid):

    if not uuid:
        return True

    if not wait_for_tms_tagent():
        return False

    res = True

    ret = subprocess.call([log_agent, 'push-dir', '--dirpath={}'.format(leapp_log_dir), '--uuid={}'.format(uuid)])
    if ret != 0:
        log("Failed to upload logs directory {}: {}".format(leapp_log_dir, ret))
        res = False
    else:
        log("Uploaded logs directory {} successfully.".format(leapp_log_dir))

    ret = subprocess.call([log_agent, 'push-dir', '--dirpath={}'.format(tmp_log_dir), '--uuid={}'.format(uuid)])
    if ret != 0:
        log("Failed to upload logs directory {}: {}".format(tmp_log_dir, ret.stderr))
        res = False
    else:
        log("Uploaded logs directory {} successfully.".format(tmp_log_dir))

    return res

def upload_log(log_path, uuid, stage):

    if not uuid:
        return True

    if not wait_for_tms_tagent():
        return True

    res = True

    ret = subprocess.call([log_agent, 'push-file', '--filepath={}'.format(log_path), '--uuid={}'.format(uuid), '--stage={}'.format(stage)])
    if ret != 0:
        log("Failed to upload log file {}: {}".format(ret.stderr))
        
        res = False

    return res


def log_agent_mark(uuid, stage, status):
    if not uuid:
        return

    command = 'echo "ok" | {} pipe-stdin --stage={} --status={} --uuid={}'.format(log_agent, stage, status, uuid)
    result = subprocess.call(command, shell=True)


def log_agent_error(error_msg, uuid, stage):
    command = 'echo "{}" | {} pipe-stdin --stage={} --status=error --uuid={}'.format(error_msg, log_agent, stage, uuid)
    result = subprocess.call(command, shell=True)


def arg_parse(input_args):
    parser = argparse.ArgumentParser(description="OC Upgrade Tool")
    parser.add_argument('--precheck', action='store_true', help='Execute precheck process')
    parser.add_argument('--upgrade', action='store_true', help='Execute upgrade process')
    parser.add_argument('--auto', action='store_true', help='Automatically execute upgrade/postcheck/clean if precheck succeeds')
    parser.add_argument('--postcheck', action='store_true', help='Execute postcheck process')
    parser.add_argument('--userapp', type=str, help='User application for precheck')
    parser.add_argument('--serverip', type=str, help='Server IP for precheck')
    parser.add_argument('--userscript', type=str, help='User script for postcheck, support comma-separated commands and entire script path', metavar='[script or CMD list]')
    parser.add_argument('--backup-dir', type=str, help='Backup directory for upgrade')
    parser.add_argument('--exclude-dirs', nargs='+', help='Directories to exclude from upgrade')
    parser.add_argument('--force', action='store_true', help='Force upgrade even if precheck fails(exclude leapp preupgrade)')
    parser.add_argument('--clean', action='store_true', help='Execute clean process')
    parser.add_argument('--uuid', type=str, help='Upload logs to Tmanager with UUID', metavar='UUID')
    parser.add_argument('--target', type=str, help='Target version for upgrade')
    args = parser.parse_args(input_args)

    exit_code = 0

    error_msg = ""
    current_stage = "precheck"

    if args.target and args.target not in supported_versions:
        exit_code = 1
        error_msg = "Unsupported target version: {}".format(args.target)
        log_agent_error(error_msg, args.uuid, current_stage)
        parser.error(error_msg)

    elif args.precheck:
        # mark the start of precheck
        log_agent_mark(args.uuid, "precheck", "begin")
        current_stage = "precheck"

        if not args.target:
            exit_code = 1
            error_msg = "Missing target version for precheck"
            log_agent_error(error_msg, args.uuid, current_stage)
            parser.error(error_msg)

        if not prepare_env(args.target):
            error_msg = "Failed to prepare environment."
            log(error_msg)
            upload_logs_dir(args.uuid)

            log_agent_error(error_msg, args.uuid, current_stage)
            exit_code = 1
            return current_stage, exit_code, args.uuid

        if not precheck(args.backup_dir, args.exclude_dirs, args.userapp, args.serverip, args.target, args.force, args.uuid):
            log("Precheck failed.")
            upload_logs_dir(args.uuid)

            error_msg = "Precheck failed."

            log_agent_error(error_msg, args.uuid, current_stage)
            exit_code = 1
            return current_stage, exit_code, args.uuid

        if args.auto:
            # mark the start of backup
            log_agent_mark(args.uuid, "backup", "begin")
            current_stage = "backup"

            if not backup("backup", args.backup_dir, args.exclude_dirs, args.uuid):
                log("Backup failed. Upgrade aborted.")
                error_msg = "Backup failed."

                log_agent_error(error_msg, args.uuid, current_stage)
                exit_code = 1
                return current_stage, exit_code, args.uuid

            # mark the start of upgrade
            log_agent_mark(args.uuid, "upgrade", "begin")
            current_stage = "upgrade"

            if not create_systemd_service(args.userscript, args.backup_dir, args.uuid, args.target):
                log("Failed to create systemd service.")
                error_msg = "Failed to create systemd service."
                exit_code = 1

            if not upgrade(args.target, args.uuid):
                log("Upgrade failed.")
                error_msg = "Upgrade failed."
                exit_code = 1

            if exit_code == 1:
                upload_logs_dir(args.uuid)
        else:
            if clean_rpms() and clean_backup(args.backup_dir) and upload_logs_dir(args.uuid):
                # 清理、上传日志失败，不清理日志
                clean_logs()

    elif args.upgrade:
        if not args.backup_dir or not args.exclude_dirs:
            parser.error("Missing required arguments for upgrade")

        # mark the start of upgrade
        log_agent_mark(args.uuid, "upgrade", "begin")
        current_stage = "upgrade"

        if not upgrade(args.target, args.uuid):
            log("Upgrade failed.")
            error_msg = "Upgrade failed."
            exit_code = 1

        if exit_code == 1:
            upload_logs_dir(args.uuid)

    elif args.postcheck:
        if not args.userscript or not args.backup_dir:
            parser.error("Missing required arguments for postcheck")

        # mark the start of precheck
        log_agent_mark(args.uuid, "postcheck", "begin")
        current_stage = "postcheck"

        ret = postcheck(args.userscript, args.uuid)
        if not ret:
            log("Postcheck failed.")
            error_msg = "Postcheck failed."
            exit_code = 1
        
        if clean_rpms() and clean_backup(args.backup_dir) and upload_logs_dir(args.uuid):
            # 清理、上传日志失败，不清理日志
            clean_logs()

    elif args.clean:
        if not args.backup_dir:
            parser.error("Missing required arguments for clean")

        clean_rpms()
        clean_backup(args.backup_dir)
        clean_logs()
    
    if exit_code == 1:
        log_agent_error(error_msg, args.uuid, current_stage)
    
    return current_stage, exit_code, args.uuid

def main():
    stage, retcode, uuid = arg_parse(sys.argv[1:])

    log("Finished at {} stage, ret: {}".format(stage, retcode))

    # mark that all processes are completed
    log_agent_mark(uuid, "completed", "finish")
    sys.exit(retcode)

if __name__ == "__main__":
    main()
