#!/usr/bin/bash

# 默认值
DEFAULT_BACKUP_DIR="/mnt"
DEFAULT_SPACE_RESERVED_FOR_LEAPP=15

# 初始化变量
BACKUP_DIR=$DEFAULT_BACKUP_DIR

# 帮助信息
usage() {
    echo "Usage: $0 -a [backup|restore|check] -d BACKUP_DIR [-e EXCLUDE_DIRS]"
    exit 1
}

# 解析选项
while getopts ":a:d:e:" opt; do
    case ${opt} in
        a )
            ACTION=$OPTARG
            ;;
        d )
            BACKUP_DIR=$OPTARG
            ;;
        e )
            EXCLUDE_DIRS=$OPTARG
            ;;
        \? )
            usage
            ;;
    esac
done

# 检查必需的参数
if [ -z "$ACTION" ] || [ -z "$BACKUP_DIR" ]; then
    usage
fi

# 检查目标目录是否存在
if [ ! -d "$BACKUP_DIR" ]; then
    echo "Backup directory $BACKUP_DIR does not exist. Creating it now..."
    mkdir -p "$BACKUP_DIR"
    if [ $? -ne 0 ]; then
        echo "Failed to create directory $BACKUP_DIR."
        exit 1
    fi
else
    # 检查目录是否为空
    if [ "$(ls -A $BACKUP_DIR)" ]; then
        echo "Backup directory $BACKUP_DIR is not empty."
        exit 1
    else
        echo "Backup directory $BACKUP_DIR is empty. Continuing..."
    fi
fi

EXCLUDE_PARAMS=""
if [ -n "$EXCLUDE_DIRS" ]; then
    # 生成 rsync 的 --exclude 参数
    IFS=',' read -r -a EXCLUDE_ARRAY <<< "$EXCLUDE_DIRS"
    for EXCLUDE_DIR in "${EXCLUDE_ARRAY[@]}"; do
        EXCLUDE_PARAMS+="--exclude=$EXCLUDE_DIR "
    done
fi

EXCLUDE_PARAMS+="--exclude=$BACKUP_DIR "

# 检查可用空间
check_space() {
    REQUIRED_SPACE=$(du -sk --exclude="$BACKUP_DIR" $EXCLUDE_PARAMS / | awk '{print $1}')
    AVAILABLE_SPACE=$(df -k --output=avail "$BACKUP_DIR" | tail -n1)

    echo "Required space: $REQUIRED_SPACE KB"
    echo "Available space: $AVAILABLE_SPACE KB"

    if [ "$AVAILABLE_SPACE" -lt "$REQUIRED_SPACE" ]; then
        echo "Not enough space in $BACKUP_DIR. Required: $REQUIRED_SPACE KB, Available: $AVAILABLE_SPACE KB."
        exit 1
    else
        echo "There is enough space in $BACKUP_DIR."
    fi
}

# 备份功能
backup() {
    REQUIRED_SPACE=$(du -sk --exclude="$BACKUP_DIR" $EXCLUDE_PARAMS / | awk '{print $1}')
    BACKUP_SPACE_GB=$((REQUIRED_SPACE / 1024 / 1024 + DEFAULT_SPACE_RESERVED_FOR_LEAPP))

    check_space

    echo "Starting backup..."
    rsync -aAXHv --delete --exclude='/dev' --exclude='/proc' --exclude='/sys' --exclude='/tmp' --exclude='/run' --exclude='/lost+found' $EXCLUDE_PARAMS / "$BACKUP_DIR"
    echo "Backup completed."
}

# 恢复功能
restore() {
    echo "Starting restore..."
    rsync -aAXHv --delete --exclude='/dev' --exclude='/proc' --exclude='/sys' --exclude='/tmp' --exclude='/run' --exclude='/lost+found' $EXCLUDE_PARAMS "$BACKUP_DIR" /
    echo "Restore completed."
}

# 执行操作
case $ACTION in
    backup)
        backup
        ;;
    restore)
        restore
        ;;
    check)
        check_space
        ;;
    *)
        echo "Invalid action: $ACTION. Use 'backup', 'restore', or 'check'."
        exit 1
        ;;
esac