#!/usr/bin/bash

# 定义要检查的关键服务列表和必须运行的服务列表
services=(
    "sshd"
    "NetworkManager"
    "libvirtd"
    "firewalld"
    "crond"
    "docker"
    "httpd"  # Apache HTTP Server
    "mysqld" # MySQL Server
    "nginx"  # Nginx
    "systemd-networkd" # Network Service
    "systemd-resolved" # DNS Service
    "systemd-timesyncd" # Time Synchronization Service
    "smbd" # SMB Server
    "nfs-server" # NFS Server
    "rpcbind" # RPC Bind
    "systemd-logind" # Login Service
    "dbus"
)

# 定义必须运行的服务列表
must_run_services=(
    "sshd"
    "systemd-logind"
    "dbus"
    "NetworkManager"
)

# 服务状态码含义
# 0 - 服务正在运行
# 3 - 服务异常（已安装）
# 4 - 服务未安装
# 5 - 应该正常运行的服务状态异常

# 检查必须运行的服务是否在运行
check_must_run_services() {
    local error_flag=0
    if [ -f $status_file ]; then
        for service in "${must_run_services[@]}"; do
            systemctl status $service > /dev/null 2>&1
            status=$?
            if [ $status -eq 3 -o $status -eq 4 ]; then
                if grep -q "^$service" $status_file; then
                    sed -i "s/^$service.*/$service 5/" $status_file
                fi
            fi
        done
    fi
    return $error_flag
}

# 记录服务状态
record_status() {
    # 初始化退出状态码
    exit_code=0

    # 清空状态文件
    > $status_file

    # 记录每个服务的状态
    for service in "${services[@]}"; do
        systemctl status $service > /dev/null 2>&1
        status=$?
        echo "$service $status" >> $status_file
    done

    # 检查必须运行的服务
    check_must_run_services
    if [ $? -ne 0 ]; then
        exit_code=1
    fi

    echo "Service statuses have been recorded in $status_file"

    exit $exit_code
}

# 对比服务状态
compare_status() {
    # 初始化退出状态码
    exit_code=0

    # 检查状态文件是否存在
    if [ ! -f $status_file ]; then
        echo "Status file $status_file does not exist. Please run the script with 'record' option first."
        exit 1
    fi

    # 检查必须运行的服务
    check_must_run_services
    if [ $? -ne 0 ]; then
        exit_code=1
    fi

    # 读取之前记录的状态
    declare -A previous_status
    while read -r line; do
        service=$(echo $line | awk '{print $1}')
        status=$(echo $line | awk '{print $2}')
        previous_status[$service]=$status
    done < $status_file

    # 对比每个服务的当前状态和之前记录的状态
    for service in "${services[@]}"; do
        systemctl status $service > /dev/null 2>&1
        current_status=$?
        
        # 如果当前服务在 must_run_services 列表中，并且状态是 3 或 4，则将其状态改为 5
        if [[ " ${must_run_services[@]} " =~ " ${service} " ]]; then
            if [ $current_status -eq 3 ] || [ $current_status -eq 4 ]; then
                current_status=5
            fi
        fi

        previous_status=${previous_status[$service]}

        if [ $current_status -ne $previous_status ]; then 
            echo "Status of $service has changed: previous=$previous_status, current=$current_status"
            exit_code=1
        else
            echo "Status of $service is unchanged: $current_status"
        fi

        #if [ $current_status -eq 5 ]; then
        #    echo "$service should be running, which is must_run service"
        #    exit_code=1
        #fi
    done

    # 退出脚本时返回相应的状态码
    exit $exit_code
}

# 清理状态文件
clean_status() {
    if [ -f $status_file ]; then
        rm $status_file
        echo "Status file $status_file has been removed."
    else
        echo "Status file $status_file does not exist."
    fi
}

# 检查传递的参数
if [ "$#" -ne 2 ]; then
    echo "Usage: $0 {record|compare|clean} <status_file>"
    exit 1
fi

# 获取命令行参数
action=$1
status_file=$2

# 根据参数执行相应的功能
case "$action" in
    record)
        record_status
        ;;
    compare)
        compare_status
        ;;
    clean)
        clean_status
        ;;
    *)
        echo "Usage: $0 {record|compare|clean} <status_file>"
        exit 1
        ;;
esac